import logging
import socketserver
import threading
from socketserver import TCPServer
from typing import Optional, Type

from steamship import Steamship, SteamshipError, TaskState
from steamship.cli.local_server.handler import create_safe_handler, make_handler
from steamship.invocable import (
    Invocable,
    InvocableRequest,
    Invocation,
    InvocationContext,
    LoggingConfig,
)


class ThreadedTCPServer(socketserver.ThreadingMixIn, socketserver.TCPServer):
    pass


class SteamshipHTTPServer:
    """A simple HTTP Server that wraps an invocable (package or plugin).

    To use, call:

       server = SteamshipHTTPServer(invocable)
       server.start()

    To shut down, call:

       server.stop()

    """

    invocable: Type[Invocable]
    base_url: str
    port: int
    server: TCPServer
    default_api_key: Optional[str] = (None,)
    invocable_handle: str = (None,)
    invocable_version_handle: str = (None,)
    invocable_instance_handle: str = (None,)

    def __init__(
        self,
        invocable: Type[Invocable],
        base_url: Optional[str] = None,
        port: int = 8080,
        invocable_handle: str = None,
        invocable_version_handle: str = None,
        invocable_instance_handle: str = None,
        config: dict = None,
        workspace: str = None,
    ):
        self.invocable = invocable
        self.port = port
        self.base_url = base_url or f"http://localhost:{self.port}"
        self.invocable_handle = invocable_handle
        self.invocable_version_handle = invocable_version_handle
        self.invocable_instance_handle = invocable_instance_handle
        self.config = config
        self.workspace = workspace

    def start(self):
        """Start the server."""

        handler = make_handler(
            self.invocable,
            base_url=self.base_url,
            invocable_handle=self.invocable_handle,
            invocable_version_handle=self.invocable_version_handle,
            invocable_instance_handle=self.invocable_instance_handle,
            config=self.config,
            workspace=self.workspace,
        )

        self.server = ThreadedTCPServer(("", self.port), handler)
        self.server.allow_reuse_address = True

        # Start a thread with the server -- that thread will then start one
        # more thread for each request
        self.server_thread = threading.Thread(target=self.server.serve_forever)
        # Exit the server thread when the main thread terminates
        self.server_thread.daemon = True
        self.server_thread.start()
        logging.info(f"Started local server thread on port {self.port}.")

        # Call the __dir__ method
        context = InvocationContext(invocable_url=f"{self.base_url}/")

        invocation = Invocation(
            http_verb="POST", invocation_path="__dir__", arguments={}, config=self.config
        )
        event = InvocableRequest(
            client_config=Steamship(workspace=self.workspace).config,
            invocation=invocation,
            logging_config=LoggingConfig(logging_host=None, logging_port=None),
            invocation_context=context,
        )
        handler = create_safe_handler(self.invocable)
        resp = handler(event.dict(by_alias=True), context)
        state = None
        try:
            state = resp.get("status", {}).get("state", None)
        except BaseException:
            state = "succeeded"

        if state == TaskState.failed:
            raise SteamshipError(
                message=resp.get("status", {}).get("statusMessage", "Unable to start project")
            )

    def stop(self):
        """Stop the server.

        Note: This has to be called from a different thread or else it will deadlock.
        """
        _server = self.server
        _server.shutdown()



import logging


def chunk_text(text: str, chunk_size: int = 200, chunk_overlap: int = 50):
    """Chunk text for embedding and insertion into an embedding index."""
    if chunk_size < 1:
        logging.warning(f"chunk_size was f{chunk_size}. Setting to 200")
        chunk_size = 200

    if chunk_overlap < 0:
        logging.warning(f"chunk_overlap was f{chunk_overlap}. Setting to 0")
        chunk_overlap = 0

    if chunk_overlap > chunk_size:
        logging.warning(f"chunk_size was f{chunk_size}. Setting to chunk_size - 1 of {chunk_size}")
        chunk_overlap = chunk_size - 1 if chunk_size > 1 else 1

    step_size = chunk_size - chunk_overlap

    for i in range(0, len(text), step_size):
        yield text[i : i + chunk_size]



import pytest

from steamship import Steamship
from steamship.agents.schema.chathistory import ChatHistory
from steamship.data.tags.tag_constants import ChatTag, TagKind
from steamship.data.tags.tag_utils import get_tag


@pytest.mark.usefixtures("client")
def test_chat_history_index_init(client: Steamship):

    context_keys = {}

    chat_history = ChatHistory.get_or_create(client, context_keys)
    assert chat_history.embedding_index is not None
    assert chat_history.embedding_index.handle is not None
    assert chat_history.embedding_index.embedder.id is not None

    # It should be marked as a Chat History
    assert get_tag(chat_history.tags, kind=TagKind.CHAT, name=ChatTag.HISTORY)

    chat_history_fetched = ChatHistory.get_or_create(client, context_keys)
    assert chat_history_fetched.embedding_index is not None
    assert chat_history.embedding_index.handle == chat_history_fetched.embedding_index.handle
    assert (
        chat_history.embedding_index.embedder.id == chat_history_fetched.embedding_index.embedder.id
    )

    # It should be marked as a Chat History
    assert get_tag(chat_history_fetched.tags, kind=TagKind.CHAT, name=ChatTag.HISTORY)

    block = chat_history.append_user_message(text="Hi. My birthday is today")
    chunk_tag = [tag for tag in block.tags if tag.name == ChatTag.CHUNK][0]
    assert chunk_tag.start_idx == 0
    assert chunk_tag.end_idx == 24

    # It should be marked as a Chat Message
    assert get_tag(block.tags, kind=TagKind.CHAT, name=ChatTag.MESSAGE)

    chat_history.append_user_message("I like bananas")
    chat_history.append_user_message("I also like shiny cars")
    chat_history.append_user_message("And I like programming agents")

    vehicle_search = chat_history.search("vehicle").wait()
    assert vehicle_search.items[0].tag.text == "I also like shiny cars"

    coding_search = chat_history.search("coding").wait()
    assert coding_search.items[0].tag.text == "And I like programming agents"



from abc import ABC


class PackageMixin(ABC):
    def instance_init(self):
        """The instance init method will be called ONCE by the engine when a new instance of a package or plugin has been created. By default, this does nothing."""
        pass



from steamship.agents.llms.openai import OpenAI
from steamship.agents.react import ReACTAgent
from steamship.agents.schema.message_selectors import MessageWindowMessageSelector
from steamship.agents.service.agent_service import AgentService
from steamship.agents.tools.image_generation import DalleTool
from steamship.agents.tools.search import SearchTool
from steamship.utils.repl import AgentREPL


class MyAssistant(AgentService):
    """MyAssistant is an example AgentService that exposes a single test endpoint
    for trying out Agent-based invocations. It is configured with two simple Tools
    to provide an overview of the types of tasks it can accomplish (here, search
    and image generation)."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)
        self.set_default_agent(
            ReACTAgent(
                tools=[
                    SearchTool(),
                    DalleTool(),
                ],
                llm=OpenAI(
                    self.client, temperature=0, model_name="gpt-4-0314"
                ),  # MUST PIN to older model
                message_selector=MessageWindowMessageSelector(k=2),
            )
        )


if __name__ == "__main__":
    # AgentREPL provides a mechanism for local execution of an AgentService method.
    # This is used for simplified debugging as agents and tools are developed and
    # added.
    AgentREPL(MyAssistant, agent_package_config={}).run()



"""
THIS SHOULD BE KEPT IN SYNC WITH THE DOCS: packages/adding-configuration.rst.
"""
from typing import Any, Dict, Type

from steamship import Steamship
from steamship.invocable import Config, InvocationContext, PackageService, post


class ConfigurableFavoritesPackage(PackageService):
    class FavoritesConfig(Config):
        """Configuration required to instantiate this package."""

        name: str
        favorite_color: str
        lucky_number: float
        favorite_true_false_value: bool

    def __init__(
        self, client: Steamship, config: Dict[str, Any] = None, context: InvocationContext = None
    ):
        # The superclass init method turns the config param (a Dict)
        # into the self.config object (here, a FavoritesConfig)
        super().__init__(client, config)

    # The config_cls method allows your package to return a class
    # that defines its required configuration.
    # See Developer Reference -> Accepting Configuration
    # for more details. This package uses a few configuration fields
    # to record the package user's favorite things.
    @classmethod
    def config_cls(cls) -> Type[Config]:
        return cls.FavoritesConfig

    # This method defines the package user's endpoint for adding content
    # The @post annotation automatically makes the method available as
    # an HTTP Post request. The name in the annotation defines the HTTP
    # route suffix, see Packages -> Package Project Structure.
    @post("my_faves")
    def my_faves(self) -> str:
        """Return the user's favorites, in a string."""

        return f"""
        Hey {self.config.name}!
        I can remind you of your favorites.
        Your favorite color is {self.config.favorite_color}.
        Your lucky number is {self.config.lucky_number}.
        Your favorite true/false value is {self.config.favorite_true_false_value}.
        Wow, mine too!
        """



import pytest

from steamship import Steamship
from steamship.invocable.mixins.indexer_mixin import IndexerMixin


@pytest.mark.usefixtures("client")
def test_indexer_mixin_and_qa_tool(client: Steamship):
    """Tests that we can inspect the package and mixin routes"""
    indexer = IndexerMixin(client)
    assert indexer.index_text("Mario was a very fun game.")
    res = indexer.search_index("Mario")
    assert res.items
    assert len(res.items) == 1
    item = res.items[0]
    assert item.tag.text == "Mario was a very fun game."
    assert item.tag.value == {}

    assert indexer.index_text("Sonic was also pretty good.", metadata={"sega": True})
    res = indexer.search_index("Sonic")
    assert res.items
    assert len(res.items) == 2
    item = res.items[0]
    assert item.tag.text == "Sonic was also pretty good."
    assert item.tag.value
    assert item.tag.value.get("sega", True)



from .prompt_database_question_answerer import PromptDatabaseQATool
from .vector_search_learner_tool import VectorSearchLearnerTool
from .vector_search_qa_tool import VectorSearchQATool

__all__ = ["PromptDatabaseQATool", "VectorSearchQATool", "VectorSearchLearnerTool"]



import pytest

from steamship import File, Steamship
from steamship.invocable import PackageService
from steamship.invocable.mixins.indexer_pipeline_mixin import IndexerPipelineMixin


@pytest.mark.usefixtures("client")
def test_set_file_status(client: Steamship):
    """Tests that we can inspect the package and mixin routes"""
    file = File.create(client, content="hi")
    invocable = PackageService()
    pipeline = IndexerPipelineMixin(client, invocable)
    pipeline.set_file_status(file_id=file.id, status="FOO")
    file = file.refresh()

    assert len(file.tags) == 1
    tag = file.tags[0]
    assert tag.kind == "status"
    assert tag.name == "FOO"



from abc import ABC, abstractmethod
from typing import List

import tiktoken
from pydantic.main import BaseModel

from steamship import Block
from steamship.data.tags.tag_constants import RoleTag


class MessageSelector(BaseModel, ABC):
    @abstractmethod
    def get_messages(self, messages: List[Block]) -> List[Block]:
        pass


class NoMessages(MessageSelector):
    def get_messages(self, messages: List[Block]) -> List[Block]:
        return []


def is_user_message(block: Block) -> bool:
    role = block.chat_role
    return role == RoleTag.USER


def is_assistant_message(block: Block) -> bool:
    role = block.chat_role
    return role == RoleTag.ASSISTANT


class MessageWindowMessageSelector(MessageSelector):
    k: int

    def get_messages(self, messages: List[Block]) -> List[Block]:
        msgs = messages[:]
        msgs.pop()  # don't add the current prompt to the memory
        if len(msgs) <= (self.k * 2):
            return msgs

        selected_msgs = []
        limit = self.k * 2
        scope = msgs[len(messages) - limit :]
        for block in scope:
            if is_user_message(block) or is_assistant_message(block):
                selected_msgs.append(block)

        return selected_msgs


def tokens(block: Block) -> int:
    enc = tiktoken.get_encoding("p50k_base")
    tokenized_text = enc.encode(block.text)
    return len(tokenized_text)


class TokenWindowMessageSelector(MessageSelector):
    max_tokens: int

    def get_messages(self, messages: List[Block]) -> List[Block]:
        selected_messages = []
        current_tokens = 0

        msgs = messages[:]
        msgs.pop()  # don't add the current prompt to the memory
        for block in reversed(msgs):
            if block.chat_role != RoleTag.SYSTEM and current_tokens < self.max_tokens:
                block_tokens = tokens(block)
                if block_tokens + current_tokens < self.max_tokens:
                    selected_messages.append(block)
                    current_tokens += block_tokens

        return reversed(selected_messages)



from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.mixins.transports.steamship_widget import SteamshipWidgetTransport
from steamship.agents.service.agent_service import AgentService
from steamship.utils.repl import AgentREPL


class ChatGpt(AgentService):
    """Minimal implementation of an Agent that just chats with you via the web."""

    def __init__(self, **kwargs):
        super().__init__(**kwargs)

        # The agent's planner is responsible for making decisions about what to do for a given input.
        self.set_default_agent(
            FunctionsBasedAgent(
                tools=[],
                llm=ChatOpenAI(self.client, model_name="gpt-4"),
            )
        )

        # This Mixin provides HTTP endpoints that connects this agent to a web client
        self.add_mixin(SteamshipWidgetTransport(client=self.client, agent_service=self))


if __name__ == "__main__":
    AgentREPL(
        ChatGpt,
        agent_package_config={"botToken": "not-a-real-token-for-local-testing"},
    ).run()



from enum import Enum
from typing import List, Mapping, Optional

from pydantic.main import BaseModel


class JSONType(str, Enum):
    string = "string"
    number = "number"
    integer = "integer"
    object = "object"
    array = "array"
    boolean = "boolean"
    null = "null"


class FunctionProperty(BaseModel):
    """Schema for an individual parameter used in an OpenAI function."""

    type: JSONType = JSONType.object
    """Type of the property. Defaults to object."""

    description: str
    """Description of the property. Should include format instructions."""


class FunctionParameters(BaseModel):
    """Schema for the description of how to invoke an OpenAI function."""

    type: JSONType = JSONType.object
    """Type of this object. DO NOT MODIFY."""

    properties: Mapping[str, FunctionProperty]
    """Map of param names to their types and description"""

    required: Optional[List[str]] = []
    """List of required parameter names."""


class OpenAIFunction(BaseModel):
    """Schema for an OpenAI function that can be used in prompting."""

    name: str
    """Name of the function.  This will appear in LLM response."""

    description: str
    """Purpose of function. Should describe expected output (and format)."""

    parameters: FunctionParameters
    """Specifies how the function should be called."""



from typing import List

from steamship import Block
from steamship.agents.functional.output_parser import FunctionsBasedOutputParser
from steamship.agents.schema import Action, AgentContext, ChatAgent, ChatLLM, Tool
from steamship.data.tags.tag_constants import RoleTag


class FunctionsBasedAgent(ChatAgent):
    """Selects actions for AgentService based on OpenAI Function style LLM Prompting."""

    PROMPT = """You are a helpful AI assistant.

NOTE: Some functions return images, video, and audio files. These multimedia files will be represented in messages as
UUIDs for Steamship Blocks. When responding directly to a user, you SHOULD print the Steamship Blocks for the images,
video, or audio as follows: `Block(UUID for the block)`.

Example response for a request that generated an image:
Here is the image you requested: Block(288A2CA1-4753-4298-9716-53C1E42B726B).

Only use the functions you have been provided with."""

    def __init__(self, tools: List[Tool], llm: ChatLLM, **kwargs):
        super().__init__(
            output_parser=FunctionsBasedOutputParser(tools=tools), llm=llm, tools=tools, **kwargs
        )

    def build_chat_history_for_tool(self, context: AgentContext) -> List[Block]:
        messages: List[Block] = []

        # get system message
        system_message = Block(text=self.PROMPT)
        system_message.set_chat_role(RoleTag.SYSTEM)
        messages.append(system_message)

        messages_from_memory = []
        # get prior conversations
        if context.chat_history.is_searchable():
            messages_from_memory.extend(
                context.chat_history.search(context.chat_history.last_user_message.text, k=3)
                .wait()
                .to_ranked_blocks()
            )

            # TODO(dougreid): we need a way to threshold message inclusion, especially for small contexts

            # remove the actual prompt from the semantic search (it will be an exact match)
            messages_from_memory = [
                msg
                for msg in messages_from_memory
                if msg.id != context.chat_history.last_user_message.id
            ]

        # get most recent context
        messages_from_memory.extend(context.chat_history.select_messages(self.message_selector))

        # de-dupe the messages from memory
        ids = []
        for msg in messages_from_memory:
            if msg.id not in ids:
                messages.append(msg)
                ids.append(msg.id)

        # TODO(dougreid): sort by dates? we SHOULD ensure ordering, given semantic search

        # put the user prompt in the appropriate message location
        # this should happen BEFORE any agent/assistant messages related to tool selection
        messages.append(context.chat_history.last_user_message)

        # get completed steps
        actions = context.completed_steps
        for action in actions:
            messages.extend(action.to_chat_messages())

        return messages

    def next_action(self, context: AgentContext) -> Action:
        # Build the Chat History that we'll provide as input to the action
        messages = self.build_chat_history_for_tool(context)

        # Run the default LLM on those messages
        output_blocks = self.llm.chat(messages=messages, tools=self.tools)

        return self.output_parser.parse(output_blocks[0].text, context)



from __future__ import annotations

from typing import List

from steamship.plugin.outputs.plugin_output import PluginOutput


class BlockTypePluginOutput(PluginOutput):
    block_types_to_create: List[str]



import logging
from typing import Any, Callable, Dict, List, Optional

from steamship import Block, Steamship, Tag
from steamship.agents.logging import StreamingOpts
from steamship.agents.schema.action import Action
from steamship.agents.schema.cache import ActionCache, LLMCache

Metadata = Dict[str, Any]
EmitFunc = Callable[[List[Block], Metadata], None]


class AgentContext:
    """AgentContext contains all relevant information about a particular execution of an Agent. It is used by the
    AgentService to manage execution history as well as store/retrieve information and metadata that will be used
    in the process of an agent execution.
    """

    # A steamship-assigned ID for this memory object
    @property
    def id(self) -> str:
        return self.chat_history.file.id

    metadata: Metadata
    """Allows storage of arbitrary information that may be useful for agents and tools."""

    client: Steamship
    """Provides workspace-specific utilities for the agents and tools."""

    chat_history: "ChatHistory"  # noqa: F821
    """Record of user-package messages. It records user submitted queries/prompts and the final
    agent-driven answer sent in response to those queries/prompts. It does NOT record any chat
    history related to agent execution and action selection."""

    completed_steps: List[Action]
    """Record of agent-selected Actions and their outputs. This provides an ordered look at the
    execution sequence for this context."""

    # todo: in the future, this could be a set of callbacks like onError, onComplete, ...
    emit_funcs: List[EmitFunc]
    """Called when an agent execution has completed. These provide a way for the AgentService
    to return the result of an agent execution to the package that requested the agent execution."""

    action_cache: Optional[ActionCache]
    """Caches all interations with Tools within a Context. This provides a way to avoid duplicated
    calls to Tools when within the same context."""

    llm_cache: Optional[LLMCache]
    """Caches all interations with LLMs within a Context. This provides a way to avoid duplicated
    calls to LLMs when within the same context."""

    def __init__(self, streaming_opts: Optional[StreamingOpts] = None):
        self.metadata = {}
        self.completed_steps = []
        self.emit_funcs = []
        if streaming_opts is not None:
            self._streaming_opts = streaming_opts
        else:
            self._streaming_opts = StreamingOpts()

    @staticmethod
    def get_or_create(
        client: Steamship,
        context_keys: Dict[str, str],
        tags: List[Tag] = None,
        searchable: bool = True,
        use_llm_cache: Optional[bool] = False,
        use_action_cache: Optional[bool] = False,
        streaming_opts: Optional[StreamingOpts] = None,
    ):
        from steamship.agents.schema.chathistory import ChatHistory

        if streaming_opts is None:
            streaming_opts = StreamingOpts()

        history = ChatHistory.get_or_create(client, context_keys, tags, searchable=searchable)
        context = AgentContext(streaming_opts=streaming_opts)
        context.chat_history = history
        context.client = client

        if use_action_cache:
            context.action_cache = ActionCache.get_or_create(
                client=client, context_keys=context_keys
            )
        else:
            context.action_cache = None

        if use_llm_cache:
            context.llm_cache = LLMCache.get_or_create(client=client, context_keys=context_keys)
        else:
            context.llm_cache = None

        return context

    def __enter__(self):
        from steamship.agents.schema.chathistory import ChatHistoryLoggingHandler

        if self._streaming_opts.stream_intermediate_events:
            self._chat_history_logger = ChatHistoryLoggingHandler(
                chat_history=self.chat_history, streaming_opts=self._streaming_opts
            )
            logger = logging.getLogger()
            logger.addHandler(self._chat_history_logger)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self._chat_history_logger:
            logger = logging.getLogger()
            logger.removeHandler(self._chat_history_logger)
            self._chat_history_logger = None



from typing import Any, List, Union

import pytest
from pydantic.fields import PrivateAttr
from steamship_tests import SRC_PATH
from steamship_tests.utils.deployables import deploy_package

from steamship import Block, Steamship, SteamshipError, Task
from steamship.agents.functional import FunctionsBasedAgent
from steamship.agents.llms.openai import ChatOpenAI
from steamship.agents.schema import Action, AgentContext, Tool
from steamship.agents.service.agent_service import AgentService
from steamship.data.tags.tag_constants import ChatTag, RoleTag, TagKind, TagValueKey


def _blocks_from_invoke(client: Steamship, potential_blocks) -> List[Block]:
    try:
        if isinstance(potential_blocks, list):
            return [Block(client=client, **raw) for raw in potential_blocks]
        else:
            return [Block(client=client, **potential_blocks)]
    except Exception as e:
        raise SteamshipError(f"Could not convert to blocks: {e}")


@pytest.mark.usefixtures("client")
def test_example_with_caching_service(client: Steamship):

    # TODO(dougreid): replace the example agent with fake/free/fast tools to minimize test time / costs?

    example_caching_agent_path = (
        SRC_PATH / "steamship" / "agents" / "examples" / "example_assistant_with_caching.py"
    )
    with deploy_package(client, example_caching_agent_path, wait_for_init=True) as (
        _,
        _,
        caching_agent,
    ):
        context_id = "foo-test-caching"
        context_keys = {"id": context_id}

        # attempt without caching
        blocks_json = caching_agent.invoke("prompt", prompt="draw a cat", context_id=context_id)
        assert blocks_json
        blocks = _blocks_from_invoke(client, blocks_json)
        assert len(blocks) == 2
        assert "image" in blocks[0].text
        assert blocks[1].is_image()

        # attempt with a max_actions_per_run budget of 0 (should fail!)
        assert caching_agent.invoke("get_max_actions_per_run") == 5
        caching_agent.invoke("set_max_actions_per_run", value=0)
        assert caching_agent.invoke("get_max_actions_per_run") == 0

        with pytest.raises(SteamshipError, match="budget"):
            blocks_json = caching_agent.invoke("prompt", prompt="draw a cat", context_id=context_id)

        caching_agent.invoke("set_max_actions_per_run", value=5)
        assert caching_agent.invoke("get_max_actions_per_run") == 5

        agent_context = AgentContext.get_or_create(
            client=client, context_keys=context_keys, use_llm_cache=True, use_action_cache=True
        )
        # just clear the cache to start fresh
        agent_context.llm_cache.clear()
        agent_context.action_cache.clear()

        fake_action = Action(tool="fake_tool", input=[Block(text="a cat")])
        agent_context.llm_cache.update(key=[Block(text="draw a cat")], value=fake_action)

        fake_output = [Block(text="psyche")]
        agent_context.action_cache.update(key=fake_action, value=fake_output)

        fake_finish_action = Action(
            tool="Agent-FinishAction", input=[], output=[Block(text="insert cat image here")]
        )
        agent_context.llm_cache.update(key=fake_output, value=fake_finish_action)

        # now attempt with the caching
        blocks_json = caching_agent.invoke("prompt", prompt="draw a cat", context_id=context_id)
        assert blocks_json
        blocks = _blocks_from_invoke(client, blocks_json)
        assert len(blocks) == 1
        assert "insert cat image here" == blocks[0].text


class FakeUncachableTool(Tool):

    name = "FakeUncacheableTool"
    human_description = "Fake tool"
    agent_description = "Ignored"

    _runs: int = PrivateAttr(default=0)

    def __init__(self, **kwargs):
        super().__init__(**kwargs, cacheable=False)

    @property
    def runs(self):
        return self._runs

    def run(self, tool_input: List[Block], context: AgentContext) -> Union[List[Block], Task[Any]]:
        self._runs += 1
        return [Block(text="sunny and mild")]


class FakeCachingService(AgentService):
    def __init__(self, **kwargs):
        super().__init__(**kwargs, use_llm_cache=True, use_action_cache=True)
        self.set_default_agent(
            FunctionsBasedAgent(
                tools=[FakeUncachableTool()], llm=ChatOpenAI(self.client, temperature=0)
            )
        )

    @property
    def uncachable_tool(self):
        return self.get_default_agent().tools[0]


@pytest.mark.usefixtures("client")
def test_non_cacheable_tool_actions(client: Steamship):
    context_id = "foo-test-non-cacheable"
    context_keys = {"id": context_id}

    agent_context = AgentContext.get_or_create(
        client=client, context_keys=context_keys, use_llm_cache=True, use_action_cache=True
    )

    service_under_test = FakeCachingService(client=client, config={})

    # clear the cache to start fresh
    agent_context.llm_cache.clear()
    agent_context.action_cache.clear()

    # build up cache to ensure Fake Weather Tool is called
    fake_action = Action(
        tool=service_under_test.uncachable_tool.name, input=[Block(text="weather in SF today")]
    )
    agent_context.llm_cache.update(
        key=[Block(text="what is the weather in SF today?")], value=fake_action
    )

    fake_finish_action = Action(
        tool="Agent-FinishAction", input=[], output=[Block(text="sunny and mild")]
    )
    agent_context.llm_cache.update(key=[Block(text="sunny and mild")], value=fake_finish_action)

    # run once to "fill" caches
    service_under_test.prompt(prompt="what is the weather in SF today?", context_id=context_id)
    assert service_under_test.uncachable_tool.runs == 1

    # now attempt again, and verify the Tool executed again
    service_under_test.prompt(prompt="what is the weather in SF today?", context_id=context_id)
    assert service_under_test.uncachable_tool.runs == 2, "The tool action was unexpectedly cached."


def has_status_message(blocks: List[Block], role: RoleTag) -> bool:
    for b in blocks:
        for t in b.tags:
            if (
                t.kind == TagKind.CHAT
                and t.name == ChatTag.ROLE
                and t.value.get(TagValueKey.STRING_VALUE) == role
            ):
                return True


@pytest.mark.usefixtures("client")
def test_context_logging_to_chat_history_everything(client: Steamship):
    example_agent_service_path = (
        SRC_PATH / "steamship" / "agents" / "examples" / "example_assistant.py"
    )
    with deploy_package(client, example_agent_service_path, wait_for_init=True) as (
        _,
        _,
        agent_service,
    ):
        # test for everything (should include actions, llm, and tool)
        context_id = "foo-test-logging-everything"
        context_keys = {"id": context_id}
        agent_context = AgentContext.get_or_create(client=client, context_keys=context_keys)
        chat_history = agent_context.chat_history

        agent_service.invoke(
            "prompt", prompt="who is the current president of Taiwan?", context_id=context_id
        )
        chat_history.refresh()
        assert len(chat_history.messages) != 0
        assert has_status_message(chat_history.messages, RoleTag.AGENT)
        assert has_status_message(chat_history.messages, RoleTag.LLM)
        assert has_status_message(chat_history.messages, RoleTag.TOOL)

        # test for individual bits (only include actions, llm, or tool)
        context_id = "foo-test-logging-agent-only"
        context_keys = {"id": context_id}
        agent_context = AgentContext.get_or_create(client=client, context_keys=context_keys)
        chat_history = agent_context.chat_history

        agent_service.invoke(
            "prompt",
            prompt="who is the current prime minister of Canada?",
            context_id=context_id,
            include_llm_messages=False,
            include_tool_messages=False,
        )
        chat_history.refresh()

        assert len(chat_history.messages) != 0
        assert has_status_message(chat_history.messages, RoleTag.AGENT)
        assert not has_status_message(chat_history.messages, RoleTag.LLM)
        assert not has_status_message(chat_history.messages, RoleTag.TOOL)

        context_id = "foo-test-logging-llm-only"
        context_keys = {"id": context_id}
        agent_context = AgentContext.get_or_create(client=client, context_keys=context_keys)
        chat_history = agent_context.chat_history

        agent_service.invoke(
            "prompt",
            prompt="who is the current prime minister of England?",
            context_id=context_id,
            include_agent_messages=False,
            include_tool_messages=False,
        )
        chat_history.refresh()

        assert len(chat_history.messages) != 0
        assert not has_status_message(chat_history.messages, RoleTag.AGENT)
        assert has_status_message(chat_history.messages, RoleTag.LLM)
        assert not has_status_message(chat_history.messages, RoleTag.TOOL)

        context_id = "foo-test-logging-tool-only"
        context_keys = {"id": context_id}
        agent_context = AgentContext.get_or_create(client=client, context_keys=context_keys)
        chat_history = agent_context.chat_history

        agent_service.invoke(
            "prompt",
            prompt="who is the current president of France?",
            context_id=context_id,
            include_agent_messages=False,
            include_llm_messages=False,
        )
        chat_history.refresh()

        assert len(chat_history.messages) != 0
        assert not has_status_message(chat_history.messages, RoleTag.AGENT)
        assert not has_status_message(chat_history.messages, RoleTag.LLM)
        assert has_status_message(chat_history.messages, RoleTag.TOOL)



from abc import ABC, abstractmethod
from typing import List

from steamship.data.block import Block, StreamState
from steamship.invocable import InvocableResponse, post
from steamship.invocable.plugin_service import PluginRequest, PluginService
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.inputs.raw_block_and_tag_plugin_input_with_preallocated_blocks import (
    RawBlockAndTagPluginInputWithPreallocatedBlocks,
)
from steamship.plugin.outputs.block_type_plugin_output import BlockTypePluginOutput
from steamship.plugin.outputs.stream_complete_plugin_output import StreamCompletePluginOutput

# Note!
# =====
#
# This is the PLUGIN IMPLEMENTOR's View of a Streaming Generator.
#
# If you are using the Steamship Client, you probably want steamship.client.operations.generator instead
# of this file.
#


class StreamingGenerator(
    PluginService[RawBlockAndTagPluginInputWithPreallocatedBlocks, StreamCompletePluginOutput], ABC
):
    @abstractmethod
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInputWithPreallocatedBlocks]
    ) -> InvocableResponse[StreamCompletePluginOutput]:
        raise NotImplementedError()

    @post("streamResultToBlocks")
    def run_endpoint(self, **kwargs) -> InvocableResponse[StreamCompletePluginOutput]:
        input = PluginRequest[RawBlockAndTagPluginInputWithPreallocatedBlocks].parse_obj(kwargs)
        for block in input.data.blocks:
            block.client = self.client
        for block in input.data.output_blocks:
            block.client = self.client
        return self.run(input)

    @post("determineOutputBlockTypes")
    def determine_output_block_types_endpoint(
        self, **kwargs
    ) -> InvocableResponse[BlockTypePluginOutput]:
        input = PluginRequest[RawBlockAndTagPluginInput].parse_obj(kwargs)
        for block in input.data.blocks:
            block.client = self.client
        try:
            return self.determine_output_block_types(input)
        except BaseException as e:
            # If anything goes wrong, make sure
            # we automatically abort any open streams
            # so the client can know
            self.abort_open_block_streams(input.data.blocks)
            raise e

    @abstractmethod
    def determine_output_block_types(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[BlockTypePluginOutput]:
        raise NotImplementedError()

    def abort_open_block_streams(self, blocks: List[Block]):
        for block in blocks:
            refreshed_block = Block.get(block.client, block.id)
            if refreshed_block.stream_state == StreamState.STARTED:
                refreshed_block.abort_stream()



from steamship_tests import PACKAGES_PATH
from steamship_tests.utils.deployables import deploy_package
from steamship_tests.utils.fixtures import get_steamship_client

from steamship.utils.url import Verb


def test_example_project_structure():

    client = get_steamship_client()
    demo_package_path = PACKAGES_PATH / "example_project_structure.py"

    with deploy_package(client, demo_package_path, wait_for_init=True) as (_, _, instance):
        response = instance.invoke("say_hello", verb=Verb.GET)
        assert response is not None
        assert response == "Hello, None"

        response = instance.invoke("say_hello", verb=Verb.GET, name="Testy")
        assert response is not None
        assert response == "Hello, Testy"

        response = instance.invoke("do_something")
        assert response is not None
        assert response == {"number": None}

        response = instance.invoke("do_something", number=3)
        assert response is not None
        assert response == {"number": 3}



from typing import List

from steamship import Block
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput


class RawBlockAndTagPluginInputWithPreallocatedBlocks(RawBlockAndTagPluginInput):
    output_blocks: List[Block]



from typing import Dict, List

from steamship import File, SteamshipError, Tag
from steamship.invocable import PackageService, post


class MockSlackApi(PackageService):
    WEBHOOK_TAG = "webhook"
    TEXT_MESSAGE_TAG = "text_message"
    CHAT_ID_KEY = "chat_id"
    PHOTO_MESSAGE_TAG = "photo_message"
    PHOTO_KEY = "photo"
    AUDIO_MESSAGE_TAG = "audio_message"
    AUDIO_KEY = "audio"
    VIDEO_MESSAGE_TAG = "video_message"
    VIDEO_KEY = "video"

    @staticmethod
    def generate_inbound_webhook_body(text: str) -> Dict:
        """Generates a Slack inbound message that conforms to a specific structure which will cause the test agent to respond with a specific media type."""
        return {
            "event": {
                "type": "message",
                "channel": "1",
                "blocks": [{"elements": [{"text": text}]}],
            },
        }

    def send_message(self, chat_id: str, text: str):
        File.create(
            self.client,
            blocks=[],
            tags=[Tag(kind=self.TEXT_MESSAGE_TAG, name=text, value={self.CHAT_ID_KEY: chat_id})],
        )
        return "OK"

    def send_photo(
        self, chat_id: str, photo: str
    ):  # realize str is wrong here, but seems to be what I get from the file upload to our server.
        File.create(
            self.client,
            blocks=[],
            tags=[
                Tag(
                    kind=self.PHOTO_MESSAGE_TAG,
                    value={self.CHAT_ID_KEY: chat_id, self.PHOTO_KEY: photo},
                )
            ],
        )
        return "OK"

    def send_audio(
        self, chat_id: str, audio: str
    ):  # realize str is wrong here, but seems to be what I get from the file upload to our server.
        File.create(
            self.client,
            blocks=[],
            tags=[
                Tag(
                    kind=self.AUDIO_MESSAGE_TAG,
                    value={self.CHAT_ID_KEY: chat_id, self.AUDIO_KEY: audio},
                )
            ],
        )
        return "OK"

    def send_video(
        self, chat_id: str, video: str
    ):  # realize str is wrong here, but seems to be what I get from the file upload to our server.
        File.create(
            self.client,
            blocks=[],
            tags=[
                Tag(
                    kind=self.VIDEO_MESSAGE_TAG,
                    value={self.CHAT_ID_KEY: chat_id, self.VIDEO_KEY: video},
                )
            ],
        )
        return "OK"

    @post("chat.postMessage", public=True)
    def post_message(self, channel: str, text: str, blocks: List[dict]):
        """Simulate the chat.postMessage command in the Slack API."""

        for block in blocks:
            kind = block.get("type")
            if kind == "video":
                video_url = block.get("video_url")
                return self.send_video(channel, video_url)
            elif kind == "section":
                if "Audio Message" in block.get("text", {}).get("text"):
                    txt = block.get("text", {}).get("text")
                    audio_url = txt.split("<")[1].split("|")[0]
                    return self.send_audio(channel, audio_url)
                else:
                    txt = block.get("text", {}).get("text")
                    return self.send_message(channel, txt)
            elif kind == "image":
                image_url = block.get("image_url")
                return self.send_photo(channel, image_url)
        raise SteamshipError(message="Unsure how to respond")



from typing import Optional

from steamship import File, Steamship, Task
from steamship.invocable import PackageService, post
from steamship.invocable.mixins.blockifier_mixin import BlockifierMixin
from steamship.invocable.mixins.file_importer_mixin import FileImporterMixin
from steamship.invocable.mixins.indexer_mixin import IndexerMixin
from steamship.invocable.package_mixin import PackageMixin
from steamship.utils.file_tags import update_file_status


class IndexerPipelineMixin(PackageMixin):
    """Provides a complete set of endpoints & async workflow for Document Question Answering.

    This Mixin is an async orchestrator of other mixins:
    - Importer Mixin:       to import files, e.g. YouTube videos, PDF urls
    - Blockifier Mixin:     to convert files to Blocks -- whether that's s2t or PDF parsing, etc.
    - Indexer Mixin:        to convert Steamship Files to embedded sharts

    """

    client: Steamship
    invocable: PackageService
    blockifier_mixin: BlockifierMixin
    importer_mixin: FileImporterMixin
    indexer_mixin: IndexerMixin

    def __init__(self, client: Steamship, invocable: PackageService):
        self.client = client
        self.invocable = invocable

        self.importer_mixin = FileImporterMixin(client)
        self.invocable.add_mixin(self.importer_mixin)

        self.blockifier_mixin = BlockifierMixin(client)
        self.invocable.add_mixin(self.blockifier_mixin)

        self.indexer_mixin = IndexerMixin(client)
        self.invocable.add_mixin(self.indexer_mixin)

    @post("/set_file_status")
    def set_file_status(self, file_id: str, status: str) -> bool:
        """Set the status bit of a file. Intended to be scheduled after import."""
        file = File.get(self.client, _id=file_id)
        update_file_status(self.client, file, status)
        return True

    @post("/index_url")
    def index_url(
        self,
        url: str,
        metadata: Optional[dict] = None,
        index_handle: Optional[str] = None,
        mime_type: Optional[str] = None,
    ) -> Task:
        """Load a URL into an embedding index.

        URL Types supported:
        - PDF (Text)
        - TXT and Markdown
        - YouTube (Though failure rate is high)

        Optional arguments:
        - mime_type (if it can be guessed by the Content-Type header or the URL schema)
        - index_handle (uses your default index if blank)
        - metadata (returned on embedding results for source attribution)
        """
        # Step 1: Import the URL
        file, task = self.importer_mixin.import_url_to_file_and_task(url)

        # Step 2: Blockify the File
        importer_task_id = None
        if task and task.task_id:
            importer_task_id = task.task_id

        blockify_task = self.blockifier_mixin.blockify(
            file_id=file.id, mime_type=mime_type, after_task_id=importer_task_id
        )

        # Step 3: Index the File
        _metadata = {"url": url}
        if metadata is not None:
            _metadata.update(metadata)

        index_task = self.invocable.invoke_later(
            method="index_file",
            wait_on_tasks=[blockify_task],
            arguments={"file_id": file.id, "index_handle": index_handle, "metadata": _metadata},
        )

        # Step 4: Set the File Status to 'indexed'
        self.invocable.invoke_later(
            method="set_file_status",
            wait_on_tasks=[index_task],
            arguments={
                "file_id": file.id,
                "status": "Indexed",
            },
        )

        # We return the index task instead of the file set task just to safe a few seconds.
        return index_task



from steamship import Block
from steamship.invocable import InvocableResponse
from steamship.plugin.generator import Generator
from steamship.plugin.inputs.raw_block_and_tag_plugin_input import RawBlockAndTagPluginInput
from steamship.plugin.outputs.plugin_output import UsageReport
from steamship.plugin.outputs.raw_block_and_tag_plugin_output import RawBlockAndTagPluginOutput
from steamship.plugin.request import PluginRequest


class RequestIDGenerator(Generator):
    # Just generate a block with the request ID as the text.
    def run(
        self, request: PluginRequest[RawBlockAndTagPluginInput]
    ) -> InvocableResponse[RawBlockAndTagPluginOutput]:
        return InvocableResponse(
            data=RawBlockAndTagPluginOutput(
                blocks=[Block(text=self.client.config.request_id)],
                usage=[UsageReport.run_tokens(5)],
            )
        )



from steamship import File, Steamship, SteamshipError, Tag


def update_file_status(client: Steamship, file: File, status: str) -> None:
    file = file.refresh()
    status_tags = [tag for tag in file.tags if tag.kind == "status"]
    for status_tag in status_tags:
        try:
            status_tag.client = client
            status_tag.delete()
        except SteamshipError:
            pass

    Tag.create(client, file_id=file.id, kind="status", name=status)



import abc
import contextlib
import logging
import signal
import uuid
from abc import ABC
from functools import partial
from json import JSONDecodeError
from typing import Any, Dict, List, Optional, Type, Union, cast

import requests

from steamship import Block, Steamship, Task, TaskState
from steamship.agents.logging import AgentLogging
from steamship.agents.schema import AgentContext, Tool
from steamship.agents.service.agent_service import AgentService
from steamship.data.workspace import Workspace
from steamship.invocable.dev_logging_handler import DevelopmentLoggingHandler

try:
    from termcolor import colored  # noqa: F401
except ImportError:

    def colored(text: str, color: str, **kwargs):
        print(text)


class SteamshipREPL(ABC):
    """Base class for building REPLs that facilitate running Steamship code in the IDE."""

    client: Steamship
    dev_logging_handler: DevelopmentLoggingHandler

    def __init__(
        self,
        log_level=None,
        dev_logging_handler=None,
    ):
        if not dev_logging_handler:
            self.dev_logging_handler = DevelopmentLoggingHandler.init_and_take_root()
        else:
            self.dev_logging_handler = dev_logging_handler

    def print_string(self, output: str, metadata: Optional[Dict[str, Any]] = None):
        """Print a string to console. All REPL output should ideally route through this method."""
        logging.info(
            f"{output}",
            extra={
                AgentLogging.IS_MESSAGE: True,
                AgentLogging.MESSAGE_AUTHOR: AgentLogging.AGENT,
                AgentLogging.MESSAGE_TYPE: AgentLogging.MESSAGE,
            },
        )

    def print_object(
        self, obj: Union[Task, Block, str, dict], metadata: Optional[Dict[str, Any]] = None
    ):
        """Print an object, returned by the agent or tool, to the console.

        Various epochs of the Agent SDK development have included Agents returning, to the repl: Blocks, strings, and
        Tasks. Since this is something that users can write (e.g. not controlled by the SDK) the REPL needs to handle
        all three cases in displaying output.
        """

        # A string gets printed wholesale.
        if isinstance(obj, str):
            self.print_string(obj, metadata)
            return

        # A task gets its ID printed.
        # TODO: It would be nice for this to be a link to the web UI.
        if isinstance(obj, Task):
            self.print_string(f"Task: {obj.task_id}", metadata)
            return

        # A dict is assumed to be a Block.
        if isinstance(obj, dict):
            obj = Block.parse_obj(obj)

        # A block gets handled based on what it contains.
        block = cast(Block, obj)
        if block.is_text():
            output = block.text
        elif block.url:
            output = block.url
        elif block.content_url:
            output = block.content_url
        else:
            block.set_public_data(True)
            output = f"{self.client.config.api_base}block/{block.id}/raw"
        if output:
            self.print_string(output, metadata)

    def print_object_or_objects(
        self, output: Union[List, Any], metadata: Optional[Dict[str, Any]] = None
    ):
        """Print Agent or Tool output, whether a list or a single object."""
        if isinstance(output, List):
            objects = cast(List, output)
            for obj in objects:
                self.print_object(obj, metadata)
        else:
            self.print_object(output, metadata)

    @contextlib.contextmanager
    def temporary_workspace(self) -> Steamship:
        workspace = Workspace.create(client=self.client)
        temp_client = Steamship(workspace=workspace.handle)
        yield temp_client
        workspace.delete()

    @abc.abstractmethod
    def run(self):
        raise NotImplementedError()


class ToolREPL(SteamshipREPL):
    tool: Tool
    client = Steamship

    def __init__(self, tool: Tool, client: Optional[Steamship] = None, **kwargs):
        super().__init__(**kwargs)
        self.tool = tool
        self.client = client or Steamship()

    def run_with_client(self, client: Workspace, context: Optional[AgentContext] = None):
        if context is None:
            context = AgentContext()
        context.client = client

        print(f"Starting REPL for Tool {self.tool.name}...")
        print("If you make code changes, restart this REPL. Press CTRL+C to exit at any time.\n")

        while True:
            input_text = input(colored("Input: ", "blue"))  # noqa: F821
            input_block = Block(text=input_text)
            output = self.tool.run([input_block], context=context)
            self.print_object_or_objects(output)

    def run(self):
        with self.temporary_workspace() as client:
            self.run_with_client(client)


class AgentREPL(SteamshipREPL):
    agent_class: Type[AgentService]
    agent_instance: Optional[AgentService]
    context_id: Optional[str] = None
    client = Steamship
    config = None

    def __init__(
        self,
        agent_class: Type[AgentService],
        method: Optional[str] = None,
        agent_package_config: Optional[Dict[str, Any]] = None,
        client: Optional[Steamship] = None,
        context_id: Optional[str] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.agent_class = agent_class
        self.method = method
        self.client = client or Steamship()
        self.config = agent_package_config
        self.agent_instance = None
        self.context_id = context_id or str(uuid.uuid4())

    def run_with_client(self, client: Steamship, **kwargs):
        try:
            from termcolor import colored  # noqa: F401
        except ImportError:

            def colored(text: str, color: str, **kwargs):
                print(text)

        print("Starting REPL for Agent...")
        print("If you make code changes, restart this REPL. Press CTRL+C to exit at any time.\n")

        self.agent_instance = self.agent_class(client=client, config=self.config)

        # Determine the responder, which may have been custom-supplied on the agent.
        responder = getattr(self.agent_instance, self.method or "prompt")

        while True:
            input_text = input(colored(text="Input: ", color="blue"))  # noqa: F821
            output = responder(prompt=input_text, context_id=self.context_id, **kwargs)
            self.print_object_or_objects(output)

    def run(self, dump_history_on_exit: Optional[bool] = False, **kwargs):
        with self.temporary_workspace() as client:
            if dump_history_on_exit:
                signal.signal(signal.SIGTERM, partial(self.print_history, client))
                signal.signal(signal.SIGINT, partial(self.print_history, client))
            self.run_with_client(client, **kwargs)

    def print_history(self, client: Steamship, *args, **kwargs):
        agent_ctx = AgentContext.get_or_create(
            client=client, context_keys={"id": f"{self.context_id}"}
        )
        print(f"\n\n----- Agent Chat History {agent_ctx.id} -----\n")
        history = agent_ctx.chat_history
        history.refresh()
        for block in history.messages:
            if block.is_text():
                print(f"[{block.chat_role}] {block.text}")
            else:
                print(f"[{block.chat_role}] {block.id} ({block.mime_type})")
        print("\n------------------------------\n")
        exit(0)


class HttpREPL(SteamshipREPL):
    """REPL that uses an HTTP endpoint. Best for the `ship serve` command."""

    prompt_url: Optional[AgentService]
    client = Steamship
    config = None
    context_id: Optional[str] = None

    def __init__(
        self,
        prompt_url: str,
        context_id: Optional[str] = None,
        client: Optional[Steamship] = None,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.prompt_url = prompt_url
        self.client = client or Steamship()
        self.context_id = context_id or str(uuid.uuid4())

    def run_with_client(self, client: Steamship, **kwargs):  # noqa: C901
        try:
            from termcolor import colored  # noqa: F401
        except ImportError:

            def colored(text: str, color: str):
                print(text)

        print("Starting REPL for Agent...")
        print("If you make code changes, restart this REPL. Press CTRL+C to exit at any time.\n")
        while True:
            input_text = input(colored(text="Input: ", color="blue"))  # noqa: F821
            resp = requests.post(
                self.prompt_url,
                json={"prompt": input_text, "context_id": self.context_id},
                headers={
                    "Content-Type": "application/json",
                    "Authorization": f"Bearer {self.client.config.api_key.get_secret_value()}",
                },
            )

            result = None

            if not resp.ok:
                logging.error(
                    f"Request to AgentService failed with HTTP Status {resp.status_code}."
                )
                logging.error(f"Message: {resp.text}")
            try:
                result = resp.json()
            except JSONDecodeError as ex:
                logging.exception(ex)
            except BaseException as ex:
                logging.exception(ex)

            if result:
                if isinstance(result, dict):
                    if result.get("status", {}).get("state", None) == TaskState.failed:
                        message = result.get("status", {}).get("status_message", None)
                        logging.error(
                            f"Response failed with remote error: {message or 'No message'}"
                        )
                        if suggestion := result.get("status", {}).get("status_suggestion", None):
                            logging.error(f"Suggestion: {suggestion}")
                    elif data := result.get("data", None):
                        self.print_object_or_objects(data)
                    else:
                        logging.warning(
                            "REPL interaction completed with empty data field in InvocableResponse."
                        )
                if isinstance(result, list):
                    self.print_object_or_objects(result)
                else:
                    logging.warning("Unsure how to display result:")
                    logging.warning(result)
            else:
                logging.warning("REPL interaction completed with no result to print.")

    def run(self, **kwargs):
        with self.temporary_workspace() as client:
            self.run_with_client(client, **kwargs)



from typing import List, Optional

from pydantic import BaseModel

from steamship import Block, Tag
from steamship.data import TagKind
from steamship.data.tags.tag_constants import RoleTag


class Action(BaseModel):
    """Actions represent a binding of a Tool to the inputs supplied to the tool.

    Upon completion, the Action also contains the output of the Tool given the inputs.
    """

    tool: str
    """Name of tool used by this action."""

    input: List[Block]
    """Data provided directly to the Tool (full context is passed in run)."""

    output: Optional[List[Block]]
    """Any direct output produced by the Tool."""

    is_final: bool = False
    """Whether this Action should be the final action performed in a reasoning loop.

    Setting this to True means that the executing Agent should halt any reasoning.
    """

    def to_chat_messages(self) -> List[Block]:
        tags = [
            Tag(kind=TagKind.ROLE, name=RoleTag.FUNCTION),
            Tag(kind="name", name=self.tool),
        ]
        blocks = []
        for block in self.output:
            # TODO(dougreid): should we revisit as_llm_input?  we might need only the UUID...
            blocks.append(
                Block(
                    text=block.as_llm_input(exclude_block_wrapper=True),
                    tags=tags,
                    mime_type=block.mime_type,
                )
            )

        # TODO(dougreid): revisit when have multiple output functions.
        # Current thinking: LLM will be OK with multiple function blocks in a row. NEEDS validation.
        return blocks


class FinishAction(Action):
    """Represents a final selected action in an Agent Execution."""

    tool = "Agent-FinishAction"
    input: List[Block] = []
    is_final = True
